--------------------------------------------------------------------------------
--
--   Copyright 2013-2021 Aerospike, Inc.
--
--   Licensed under the Apache License, Version 2.0 (the "License");
--   you may not use this file except in compliance with the License.
--   You may obtain a copy of the License at
--
--       http://www.apache.org/licenses/LICENSE-2.0
--
--   Unless required by applicable law or agreed to in writing, software
--   distributed under the License is distributed on an "AS IS" BASIS,
--   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
--   See the License for the specific language governing permissions and
--   limitations under the License.
--
--------------------------------------------------------------------------------

function add(rec, a, b)
    return (a or 0) + (b or 0)
end
function concat(rec, a, b)
    return (a or "") .. (b or "")
end
function add_val(rec, bin, offset)
    rec[bin] = rec[bin] + offset
	aerospike:update(rec)
end
